import torch
import math
import time
import kintera
import numpy as np
from snapy import (
    index,
    MeshBlockOptions,
    MeshBlock,
    OutputOptions,
    NetcdfOutput,
)
from kintera import (
    ThermoOptions,
    ThermoX,
    KineticsOptions,
    Kinetics,
)

if __name__ == "__main__":
    # input file
    infile = "earth.yaml"
    device = "cpu"

    # create meshblock
    op_block = MeshBlockOptions.from_yaml(infile)
    block = MeshBlock(op_block)
    block.to(torch.device(device))

    # create thermo module
    op_thermo = ThermoOptions.from_yaml(infile)
    thermo_x = ThermoX(op_thermo)
    thermo_x.to(torch.device(device))

    # create kinetics module
    op_kinet = KineticsOptions.from_yaml(infile)
    kinet = Kinetics(op_kinet)
    kinet.to(torch.device(device))

    # create output fields
    op_out = OutputOptions().file_basename("earth")
    out2 = NetcdfOutput(op_out.fid(2).variable("prim"))
    out3 = NetcdfOutput(op_out.fid(3).variable("uov"))
    out4 = NetcdfOutput(op_out.fid(4).variable("diag"))
    outs = [out2, out4]

    # set up initial condition
    w = setup_initial_condition(block, thermo_x)
    print("w = ", w[:, 0, 0, :])

    # integration
    current_time = 0.0
    count = 0
    start_time = time.time()
    interior = block.part((0, 0, 0))
    while not block.intg.stop(count, current_time):
        dt = block.max_time_step()
        u = block.buffer("hydro.eos.U")

        if count % 1 == 0:
            print(f"count = {count}, dt = {dt}, time = {current_time}")
            print("mass = ", u[interior][index.idn].sum())

            for out in outs:
                out.increment_file_number()
                out.write_output_file(block, current_time)
                out.combine_blocks()

        for stage in range(len(block.intg.stages)):
            block.forward(dt, stage)

        # evolve kinetics
        u[index.icy :] += evolve_kinetics(block, kinet, thermo_x)

        current_time += dt
        count += 1



from scipy.interpolate import interp1d
import snapy
import numpy as np

from .setup_profile import setup_profile


def find_init_params(
    block: snapy.MeshBlock,
    param: dict[str, float],
    *,
    target_T: float = 300.0,
    target_P: float = 1.0e5,
    method: str = "moist-adiabat",
    max_iter: int = 50,
    ftol: float = 1.0e-2,
    verbose: bool = True,
):
    """Find initial parameters that yield desired T and P

    Args:
        block (snapy.MeshBlock): The mesh block to set up.
        param (dict[str, float]): Initial guess parameters for the adiabat setup.
            Required keys: Ts, Ps, x<species>, grav.
        target_T (float, optional): Target temperature in Kelvin. Defaults to 300 K.
        target_P (float, optional): Target pressure in Pascals. Defaults to 1e5 Pa.
        method (str, optional): Method for the adiabat setup.
        max_iter (int, optional): Maximum number of iterations. Defaults to 50.
        ftol (float, optional): Tolerance for temperature convergence. Defaults to 1e-2 K.
        verbose (bool, optional): If True, print iteration details. Defaults to True.

    Returns:
        dict: Dictionary containing the found parameters: Ts, Ps, xH2O, xNH3, xH2S.
    """
    count = 0
    eos = block.hydro.get_eos()

    while count < max_iter:
        if verbose:
            print(f"Iteration {count+1}: Trying Ts={param['Ts']}\n")

        # setup profile
        w = setup_profile(block, param, method=method)

        # calculate temperature
        temp = eos.compute("W->T", (w,)).squeeze()

        # calculate 1D pressure
        pres = w[snapy.index.ipr, ...].squeeze()

        # temperature function
        t_func = interp1d(
            pres.log().cpu().numpy(),
            temp.log().cpu().numpy(),
            kind="linear",
            fill_value="extrapolate",
        )

        temp1 = np.exp(t_func(np.log(target_P)))
        if verbose:
            print(f"  At P={target_P:.3e} Pa, T={temp1:.3f} K (target {target_T} K)\n")
        if abs(temp1 - target_T) < ftol:
            if verbose:
                print("Converged! Found parameters:")
                for key, val in param.items():
                    print(f"  {key} = {val}")
                print(f"Matching T = {target_T} K at P = {target_P} Pa")
            return param

        # adjust Ts using a damped scaling
        param["Ts"] += (target_T - temp1) * 0.9
        count += 1

    raise RuntimeError("Failed to converge within the maximum number of iterations.")



from typing import Optional
from scipy.interpolate import interp1d

import datetime
import numpy as np
import torch
import kintera
import snapy


def write_profile(
    filename: str,
    hydro_w: torch.Tensor,
    block: snapy.MeshBlock,
    ref_pressure: float = 1.0e5,
    comment: Optional[str] = None,
) -> None:
    """
    Write an atmospheric profile to a text file.

    Args:
        filename (str): The name of the text file to write.
        hydro_w (torch.Tensor): Primitive variables of the atmosphere profile.
        block (snapy.MeshBlock): The mesh block containing the atmosphere data.
        ref_pressure (float): Reference pressure for height calculation.
        comment (Optional[str]): A comment string to write to the file.

    Returns:
        None
    """
    # useful modules
    thermo_y = block.module("hydro.eos.thermo")
    coord = block.module("hydro.coord")
    eos = block.hydro.get_eos()

    # handling mole fraction quantities
    thermo_x = kintera.ThermoX(thermo_y.options)
    thermo_x.to(dtype=hydro_w.dtype, device=hydro_w.device)

    # check dimensions of hydro_w
    species = thermo_y.options.species()
    if hydro_w.ndim != 4:
        raise ValueError("hydro_w must be a 4D tensor.")
    if hydro_w.shape[0] != 4 + len(species):
        raise ValueError("hydro_w shape does not match number of species.")
    if hydro_w.shape[1] != 1 or hydro_w.shape[2] != 1:
        raise ValueError("hydro_w must have shape (N, 1, 1, L).")

    # calculate a height grid
    pres = hydro_w[snapy.index.ipr, ...].squeeze() / 1.0e5  # Pa -> bar
    zlev_func = interp1d(
        pres.log().cpu().numpy(),
        coord.buffer("x1v").cpu().numpy(),
        kind="linear",
        fill_value="extrapolate",
    )
    zref = zlev_func(np.log(ref_pressure / 1.0e5))
    zlev = (coord.buffer("x1v") - zref) / 1.0e3  # m -> km

    # calculate temperature
    temp = eos.compute("W->T", (hydro_w,)).squeeze()

    # calculate mole fractions
    xfrac = thermo_y.compute("Y->X", (hydro_w[snapy.index.icy :, ...],)).squeeze()

    # calculate heat capacity
    conc = thermo_x.compute("TPX->V", (temp, pres * 1.0e5, xfrac))
    cpx = thermo_x.compute("TV->cp", (temp, conc)) / conc.sum(-1)

    # calculate entropy
    ens = thermo_x.compute("TPV->S", (temp, pres * 1.0e5, conc)) / conc.sum(-1)

    with open(filename, "w") as f:
        # write comments
        f.write(f"# File generated by write_profile.py\n")
        f.write(f"# Date generated: {datetime.datetime.now()}\n")
        f.write(f"# Speices units: mole fraction [ppmv]\n")
        f.write(f"# Width of the first column: 6 characters\n")
        f.write(f"# Width of other columns: 13 characters\n")
        f.write(f"# Reference pressure for z = 0: {ref_pressure} Pa\n")
        if comment:
            f.write(f"# {comment}\n\n")

        # Write header
        headers = []
        headers.append(f"# {'IDX':<4}")
        headers.append(f"{'HGT[km]':>{13}}")
        headers.append(f"{'PRE[bar]':>{13}}")
        headers.append(f"{'TEM[K]':>{13}}")
        for sp in species[1:]:
            headers.append(f"{sp.upper():>{13}}")
        headers.append(f"{'CPX[J,K,mol]':>{13}}")
        headers.append(f"{'ENS[J,K,mol]':>{13}}")

        f.write("".join(headers) + "\n")

        # Write data
        num_rows = hydro_w.shape[-1]
        for i in range(num_rows, 0, -1):
            row = []
            row.append(f"{i:<6}")
            row.append(f"{zlev[i-1]:.6g}".rjust(13))  # HGT
            row.append(f"{pres[i-1]:.6g}".rjust(13))  # PRE
            row.append(f"{temp[i-1]:.6g}".rjust(13))
            for s in range(1, len(species)):
                row.append(f"{xfrac[i-1,s]*1.e6:.6g}".rjust(13))
            row.append(f"{cpx[i-1]:.6g}".rjust(13))
            row.append(f"{ens[i-1]:.6g}".rjust(13))

            f.write("".join(row) + "\n")
    print(f"Atmosphere profile written to {filename}")



#! /usr/bin/env python3

"""
Read variables in a NetCDF file and write them to jit saved torch tensors.
Usage: python nc2pt.py input.nc output.pt
"""

from netCDF4 import Dataset
import torch


def save_tensors(tensor_map: dict[str, torch.Tensor], filename: str):
    class TensorModule(torch.nn.Module):
        def __init__(self, tensors):
            super().__init__()
            for name, tensor in tensors.items():
                self.register_buffer(name, tensor)

    module = TensorModule(tensor_map)
    scripted = torch.jit.script(module)  # Needed for LibTorch compatibility
    scripted.save(filename)


fname = "sod.out0.00019.nc"

nc = Dataset(fname, "r")
out_fname = "sod.out0.00019.pt"

data = {}
for varname in nc.variables:
    var = nc.variables[varname][:]
    if var.ndim == 4:  # (time, x1, x2, x3) -> (time, x3, x2, x1)
        data[varname] = torch.tensor(var).permute(0, 3, 2, 1).squeeze()
    elif var.ndim == 3:  # (x1, x2, x3) -> (x3, x2, x1)
        data[varname] = torch.tensor(var).permute(2, 1, 0).squeeze()
    else:
        data[varname] = torch.tensor(var).squeeze()

save_tensors(data, out_fname)
